package org.jboss.jsr299.tck.tests.context.dependent;

import java.lang.annotation.Annotation;
import java.util.Set;

import javax.enterprise.context.ContextNotActiveException;
import javax.enterprise.context.Dependent;
import javax.enterprise.context.spi.Context;
import javax.enterprise.context.spi.CreationalContext;
import javax.enterprise.inject.AnnotationLiteral;
import javax.enterprise.inject.spi.Bean;

import org.hibernate.tck.annotations.SpecAssertion;
import org.hibernate.tck.annotations.SpecAssertions;
import org.jboss.jsr299.tck.AbstractJSR299Test;
import org.jboss.jsr299.tck.impl.MockCreationalContext;
import org.jboss.testharness.impl.packaging.Artifact;
import org.jboss.testharness.impl.packaging.jsr299.BeansXml;
import org.testng.annotations.Test;

/**
 * Spec version: 20090519
 */
@Artifact
@BeansXml("beans.xml")
public class DependentContextTest extends AbstractJSR299Test
{
   
   private static final Annotation TAME_LITERAL = new AnnotationLiteral<Tame> () {};
   
   @Test(groups = { "contexts", "injection" })
   @SpecAssertion(section = "6.4", id = "a")
   public void testInstanceNotSharedBetweenInjectionPoints()
   {
      Set<Bean<Fox>> foxBeans = getBeans(Fox.class);
      assert foxBeans.size() == 1;
      Set<Bean<FoxRun>> foxRunBeans = getBeans(FoxRun.class);
      assert foxRunBeans.size() == 1;
      Bean<FoxRun> foxRunBean = foxRunBeans.iterator().next();
      CreationalContext<FoxRun> creationalContext = getCurrentManager().createCreationalContext(foxRunBean);
      FoxRun foxRun = foxRunBean.create(creationalContext);
      assert !foxRun.fox.equals(foxRun.anotherFox);
   }

   @Test(groups = { "contexts", "el" })
   @SpecAssertion(section = "6.4", id = "c")
   public void testInstanceUsedForElEvaluationNotShared() throws Exception
   {
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            Set<Bean<Fox>> foxBeans = getBeans(Fox.class);
            assert foxBeans.size() == 1;

            Fox fox1 = getCurrentConfiguration().getEl().evaluateValueExpression("#{fox}", Fox.class);
            Fox fox2 = getCurrentConfiguration().getEl().evaluateValueExpression("#{fox}", Fox.class);
            assert !fox1.equals(fox2);
         }
      }.run();
   }

   @Test(groups = { "contexts", "producerMethod" })
   @SpecAssertion(section = "6.4", id = "da")
   public void testInstanceUsedForProducerMethodNotShared() throws Exception
   {
      Bean<Tarantula> tarantulaBean = getBeans(Tarantula.class).iterator().next();
      
      CreationalContext<Tarantula> creationalContext = getCurrentManager().createCreationalContext(tarantulaBean);
      Tarantula tarantula = tarantulaBean.create(creationalContext);
      Tarantula tarantula2 = tarantulaBean.create(creationalContext);
      assert tarantula != null;
      assert tarantula2 != null;
      assert tarantula != tarantula2;
   }

   @Test(groups = { "contexts", "producerMethod" })
   @SpecAssertion(section = "6.4", id = "db")
   public void testInstanceUsedForProducerFieldNotShared() throws Exception
   {
      Bean<Tarantula> tarantulaBean = getBeans(Tarantula.class, TAME_LITERAL).iterator().next();
      CreationalContext<Tarantula> creationalContext = getCurrentManager().createCreationalContext(tarantulaBean);
      Tarantula tarantula = tarantulaBean.create(creationalContext);
      Tarantula tarantula2 = tarantulaBean.create(creationalContext);
      assert tarantula != null;
      assert tarantula2 != null;
      assert tarantula != tarantula2;
   }

   @Test(groups = { "contexts", "disposalMethod" })
   @SpecAssertion(section = "6.4", id = "dc")
   public void testInstanceUsedForDisposalMethodNotShared() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            SpiderProducer spiderProducer = getInstanceByType(SpiderProducer.class);
            Bean<Tarantula> tarantulaBean = getBeans(Tarantula.class).iterator().next();
            CreationalContext<Tarantula> creationalContext = getCurrentManager().createCreationalContext(tarantulaBean);
            Tarantula tarantula = tarantulaBean.create(creationalContext);
            assert tarantula != null;
            tarantulaBean.destroy(tarantula, creationalContext);
            assert SpiderProducer.getInstanceUsedForDisposal() != null;
            assert SpiderProducer.getInstanceUsedForDisposal() != spiderProducer;
         }

      }.run();
   }

   @Test(groups = { "contexts", "observerMethod" })
   @SpecAssertion(section = "6.4", id = "dd")
   public void testInstanceUsedForObserverMethodNotShared() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            HorseStable firstStableInstance = getInstanceByType(HorseStable.class);
            getCurrentManager().fireEvent(new HorseInStableEvent());
            assert HorseStable.getInstanceThatObservedEvent() != null;
            assert HorseStable.getInstanceThatObservedEvent() != firstStableInstance;
         }

      }.run();
   }

   @Test(groups = "contexts")
   @SpecAssertion(section = "6.4", id = "e")
   public void testContextGetWithCreationalContextReturnsNewInstance() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            Set<Bean<Fox>> foxBeans = getBeans(Fox.class);
            assert foxBeans.size() == 1;
            Bean<Fox> foxBean = foxBeans.iterator().next();
            Context context = getCurrentManager().getContext(Dependent.class);
            assert context.get(foxBean, new MockCreationalContext<Fox>()) != null;
            assert context.get(foxBean, new MockCreationalContext<Fox>()) instanceof Fox;
         }

      }.run();
   }

   @Test(groups = "contexts")
   @SpecAssertion(section = "6.4", id = "f")
   public void testContextGetWithCreateFalseReturnsNull() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            Set<Bean<Fox>> foxBeans = getBeans(Fox.class);
            assert foxBeans.size() == 1;
            Bean<Fox> foxBean = foxBeans.iterator().next();
            Context context = getCurrentManager().getContext(Dependent.class);
            assert context.get(foxBean, null) == null;
         }

      }.run();
   }

   @Test(groups = { "contexts", "ri-broken" }, expectedExceptions = ContextNotActiveException.class)
   @SpecAssertions({
      @SpecAssertion(section = "6.2", id = "ha"),
      @SpecAssertion(section = "6.4.1", id = "ga")
   })
   public void testContextIsInactive()
   {
      // The RI test harness is broken here, it just enables the dependent context blindly
      getCurrentManager().getContext(Dependent.class);
   }

   @Test(groups = { "contexts", "producerMethod" })
   @SpecAssertions({
      @SpecAssertion(section = "6.2", id = "ha"),
      @SpecAssertion(section = "6.4.1", id = "gb")
   })
   public void testContextIsActiveWhenInvokingProducerMethod()
   {
      Bean<Tarantula> tarantulaBean = getBeans(Tarantula.class).iterator().next();
      CreationalContext<Tarantula> creationalContext = getCurrentManager().createCreationalContext(tarantulaBean);
      Tarantula tarantula = tarantulaBean.create(creationalContext);
      assert tarantula != null;
      assert SpiderProducer.isDependentContextActive();
   }

   @Test(groups = { "contexts", "producerField"})
   @SpecAssertion(section = "6.4.1", id = "gc")
   public void testContextIsActiveWhenInvokingProducerField()
   {
      // Reset test class
      Tarantula.setDependentContextActive(false);
      getInstanceByType(Tarantula.class,TAME_LITERAL);
      assert Tarantula.isDependentContextActive();
   }

   @Test(groups = { "contexts", "disposalMethod" })
   @SpecAssertions({
      @SpecAssertion(section = "6.4.1", id = "gd"),
      @SpecAssertion(section = "11.1", id = "a")
   })
   public void testContextIsActiveWhenInvokingDisposalMethod()
   {
      Bean<Tarantula> tarantulaBean = getBeans(Tarantula.class).iterator().next();
      CreationalContext<Tarantula> creationalContext = getCurrentManager().createCreationalContext(tarantulaBean);
      Tarantula tarantula = tarantulaBean.create(creationalContext);
      assert tarantula != null;
      SpiderProducer.setDependentContextActive(false);
      tarantulaBean.destroy(tarantula, creationalContext);
      assert SpiderProducer.isDependentContextActive();
   }

   @Test(groups = { "contexts", "observerMethod" })
   @SpecAssertions({
      @SpecAssertion(section = "6.4.1", id = "ge"),
      @SpecAssertion(section = "6.4.1", id = "i")
   })
   public void testContextIsActiveWhenCreatingObserverMethodInstance()
   {
      getCurrentManager().fireEvent(new HorseInStableEvent());
      assert HorseStable.isDependentContextActive();
   }

   @Test(groups = { "contexts", "el" })
   @SpecAssertion(section = "6.4.1", id = "h")
   public void testContextIsActiveWhenEvaluatingElExpression() throws Exception
   {
      SensitiveFox.setManager(getCurrentManager());
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            String foxName = getCurrentConfiguration().getEl().evaluateMethodExpression("#{sensitiveFox.getName}", String.class, new Class[0], new Object[0]);
            assert foxName != null;
            assert SensitiveFox.isDependentContextActiveDuringEval();
         }
      }.run();
   }

   @Test(groups = { "contexts", "beanLifecycle" })
   @SpecAssertion(section = "6.4.1", id = "ja")
   public void testContextIsActiveDuringBeanCreation() throws Exception
   {
      SensitiveFox.setManager(getCurrentManager());
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            SensitiveFox fox1 = getInstanceByType(SensitiveFox.class);
            assert fox1 != null;
            assert fox1.isDependentContextActiveDuringCreate();
         }

      }.run();
   }

   @Test(groups = { "contexts", "injection" })
   @SpecAssertion(section = "6.4.1", id = "jc")
   public void testContextIsActiveDuringInjection()
   {
      Bean<FoxRun> foxRunBean = getBeans(FoxRun.class).iterator().next();
      FoxRun foxRun = foxRunBean.create(new MockCreationalContext<FoxRun>());
      assert foxRun.fox != null;
   }

   @Test(groups = { "contexts", "beanDestruction"})
   @SpecAssertions({
      @SpecAssertion(section = "6.4.3", id = "a"),
      @SpecAssertion(section = "6.4", id = "b")
   })
   public void testDestroyingSimpleParentDestroysDependents() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            assert getBeans(Farm.class).size() == 1;
            Bean<Farm> farmBean = getBeans(Farm.class).iterator().next();
            CreationalContext<Farm> creationalContext = getCurrentManager().createCreationalContext(farmBean);
            Farm farm = farmBean.create(creationalContext);
            farm.open();
            Stable.destroyed = false;
            Horse.destroyed = false;
            farmBean.destroy(farm, creationalContext);
            assert Stable.destroyed;
            assert Horse.destroyed;
         }

      }.run();
   }
   
   @Test(groups = { "contexts", "beanDestruction"})
   @SpecAssertions({
      @SpecAssertion(section = "6.4.3", id = "a"),
      @SpecAssertion(section = "6.4", id = "b")
   })
   public void testDestroyingSimpleParentDestroysDependentsOfSameBean() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            // Reset test class
            Fox.setDestroyed(false);
            Fox.setDestroyCount(0);
            
            assert getCurrentManager().getBeans(FoxRun.class).size() == 1;
            Bean<FoxRun> bean = getBeans(FoxRun.class).iterator().next();
            CreationalContext<FoxRun> creationalContext = getCurrentManager().createCreationalContext(bean);
            FoxRun instance = bean.create(creationalContext);
            assert instance.fox != instance.anotherFox;
            bean.destroy(instance, creationalContext);
            assert Fox.isDestroyed();
            assert Fox.getDestroyCount() == 2;
         }

      }.run();
   }

   @Test(groups = { "contexts", "el"})
   @SpecAssertion(section = "6.4.3", id = "c")
   public void testDependentsDestroyedWhenElEvaluationCompletes() throws Exception
   {
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            // Reset test class
            Fox.setDestroyed(false);
            FoxRun.setDestroyed(false);
            
            getCurrentConfiguration().getEl().evaluateValueExpression("#{foxRun}", FoxRun.class);
            assert FoxRun.isDestroyed();
            assert Fox.isDestroyed();
         }
      }.run();
   }

   @Test(groups = { "contexts", "producerMethod" })
   @SpecAssertion(section = "6.4.3", id = "da")
   public void testDependentsDestroyedWhenProducerMethodCompletes() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            // Reset the test class
            SpiderProducer.setDestroyed(false);
            Tarantula spiderInstance = getInstanceByType(Tarantula.class);
            assert spiderInstance != null;
            assert SpiderProducer.isDestroyed();
         }

      }.run();
   }

   @Test(groups = { "contexts", "producerField" })
   @SpecAssertion(section = "6.4.3", id = "db")
   public void testDependentsDestroyedWhenProducerFieldCompletes() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            // Reset the test class
            OtherSpiderProducer.setDestroyed(false);
            
            Tarantula spiderInstance = getInstanceByType(Tarantula.class,TAME_LITERAL);
            assert spiderInstance != null;
            assert OtherSpiderProducer.isDestroyed();
            
         }

      }.run();
   }

   @Test(groups = { "contexts", "disposalMethod", "ri-broken" })
   @SpecAssertion(section = "6.4.3", id = "dc")
   public void testDependentsDestroyedWhenDisposalMethodCompletes() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            Bean<Tarantula> tarantulaBean = getBeans(Tarantula.class).iterator().next();
            CreationalContext<Tarantula> creationalContext = getCurrentManager().createCreationalContext(tarantulaBean);
            Tarantula tarantula = tarantulaBean.create(creationalContext);
            assert tarantula != null;
            
            // Reset test class state
            SpiderProducer.setDestroyed(false);
            
            tarantulaBean.destroy(tarantula, creationalContext);
            assert SpiderProducer.isDestroyed();
         }

      }.run();
   }

   @Test(groups = { "contexts", "observerMethod" })
   @SpecAssertion(section = "6.4.3", id = "dd")
   public void testDependentsDestroyedWhenObserverMethodEvaluationCompletes() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            // Reset test class state...
            HorseStable.setInstanceThatObservedEvent(null);
            HorseStable.setDestroyed(false);
            
            getCurrentManager().fireEvent(new HorseInStableEvent());
            assert HorseStable.getInstanceThatObservedEvent() != null;
            assert HorseStable.isDestroyed();
         }

      }.run();
   }

}
