package org.jboss.jsr299.tck.tests.context.passivating;

import java.io.IOException;
import java.lang.annotation.Annotation;
import java.util.Set;

import javax.enterprise.context.SessionScoped;
import javax.enterprise.context.spi.Context;
import javax.enterprise.inject.AnnotationLiteral;
import javax.enterprise.inject.IllegalProductException;
import javax.enterprise.inject.spi.Bean;

import org.hibernate.tck.annotations.SpecAssertion;
import org.hibernate.tck.annotations.SpecAssertions;
import org.jboss.jsr299.tck.AbstractJSR299Test;
import org.jboss.testharness.impl.packaging.Artifact;
import org.jboss.testharness.impl.packaging.Packaging;
import org.jboss.testharness.impl.packaging.PackagingType;
import org.jboss.testharness.impl.packaging.jsr299.BeansXml;
import org.testng.annotations.Test;

/**
 * 
 * @author Nicklas Karlsson
 * @author David Allen
 * 
 * Spec version: PRD2
 */
@Artifact
@BeansXml("beans.xml")
@Packaging(PackagingType.EAR)
public class PassivatingContextTest extends AbstractJSR299Test
{
   
   private static final Annotation BIG_LITERAL = new AnnotationLiteral<Big>() {};
   
   @Test(groups = { "contexts", "passivation" })
   @SpecAssertion(section = "8.4", id = "b")
   public void testSimpleWebBeanWithSerializableImplementationClassOK()
   {
      Set<Bean<Jyvaskyla>> beans = getBeans(Jyvaskyla.class);
      assert !beans.isEmpty();
   }
   
   @Test @SpecAssertion(section="8.4", id = "ca")
   public void testInjectionOfDependentSerializableProductIntoNormalBean()
   {
      getInstanceByType(NumberConsumer.class).ping();
   }
   
   @Test @SpecAssertion(section="8.4", id = "cb")
   public void testInjectionOfDependentPrimitiveProductIntoNormalBean()
   {
      getInstanceByType(FooConsumer.class).ping();
   }

   @Test(groups = { "contexts", "passivation" })
   @SpecAssertions({
      @SpecAssertion(section = "8.4", id = "f"),
      @SpecAssertion(section = "8.4", id = "i")
   })
   public void testSimpleWebBeanDeclaringPassivatingScopeIsSerializedWhenContextIsPassivated() throws IOException, ClassNotFoundException
   {
      Kajaani instance = getInstanceByType(Kajaani.class);
      instance.setTheNumber(100);
      Context sessionContext = getCurrentManager().getContext(SessionScoped.class);
      setContextInactive(sessionContext);
      setContextActive(sessionContext);
      instance = getInstanceByType(Kajaani.class);
      assert instance.getTheNumber() == 100;
   }

   @SuppressWarnings("unchecked")
   private <T> boolean testSerialize(Bean<T> bean) throws IOException, ClassNotFoundException
   {
      getCurrentManager().addBean(bean);
      T instance = getCurrentManager().getContext(bean.getScopeType()).get(bean);
      byte[] data = serialize(instance);
      T resurrected = (T) deserialize(data);
      return resurrected.toString().equals(instance.toString());
   }

   /**
    * @throws ClassNotFoundException
    * @throws IOException
    */
   @Test(groups = { "contexts", "passivation", "ri-broken" })
   @SpecAssertion(section = "8.4", id = "j")
   public void testDependentEJBsAreSerializable() throws IOException, ClassNotFoundException
   {
      Set<Bean<Vaasa>> vaasaBeans = getBeans(Vaasa.class);
      assert vaasaBeans.size() == 1;
      assert testSerialize(vaasaBeans.iterator().next());
   }

   @Test(groups = { "contexts", "passivation" })
   @SpecAssertion(section = "8.4", id = "l")
   public void testSimpleDependentWebBeanWithNonSerializableImplementationInjectedIntoTransientFieldOK()
   {
      Set<Bean<Joensuu>> beans = getBeans(Joensuu.class);
      assert !beans.isEmpty();
   }

   @Test(groups = { "contexts", "passivation" }, expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "q")
   public void testDependentScopedProducerMethodReturnsNonSerializableObjectForInjectionIntoNonTransientFieldOfWebBeanWithPassivatingScopeFails()
   {
      getInstanceByType(Nokia_Broken.class).ping();
   }

   @Test(groups = { "contexts", "passivation" })
   @SpecAssertion(section = "8.4", id = "q")
   public void testDependentScopedProducerMethodReturnsNonSerializableObjectForInjectionIntoTransientFieldOfWebBeanWithPassivatingScopeOK()
   {
      Set<Bean<Hyvinkaa>> beans = getBeans(Hyvinkaa.class);
      assert !beans.isEmpty();
   }

   @Test(groups = { "contexts", "passivation" }, expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "r")
   public void testDependentScopedProducerMethodReturnsNonSerializableObjectForInjectionIntoConstructorParameterOfWebBeanWithPassivatingScopeFails()
   {
      getInstanceByType(Loviisa_Broken.class).ping();
   }

   @Test(groups = { "contexts", "passivation" }, expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "s")
   public void testDependentScopedProducerMethodReturnsNonSerializableObjectForInjectionIntoInitializerParameterOfWebBeanWithPassivatingScopeFails()
   {
      getInstanceByType(Kuopio_Broken.class).ping();
   }

   @Test(groups = { "contexts", "passivation" }, expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "t")
   public void testDependentScopedProducerMethodReturnsNonSerializableObjectForInjectionIntoProducerMethodParameterWithPassivatingScopeFails()
   {
      getInstanceByType(Violation2.class).ping();
   }

   @Test(groups = { "contexts", "passivation" }, expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "q")
   public void testDependentScopedProducerFieldReturnsNonSerializableObjectForInjectionIntoNonTransientFieldOfWebBeanWithPassivatingScopeFails()
   {
      getInstanceByType(Uusikaupunki_Broken.class).ping();
   }

   @Test(groups = { "contexts", "passivation" })
   @SpecAssertion(section = "8.4", id = "q")
   public void testDependentScopedProducerFieldReturnsNonSerializableObjectForInjectionIntoTransientFieldOfWebBeanWithPassivatingScopeOK()
   {
      getInstanceByType(Salo_Broken.class).ping();
   }

   @Test(groups = { "contexts", "passivation" }, expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "r")
   public void testDependentScopedProducerFieldReturnsNonSerializableObjectForInjectionIntoConstructorParameterOfWebBeanWithPassivatingScopeFails()
   {
      getInstanceByType(Loviisa_Broken.class).ping();
   }

   @Test(groups = { "contexts", "passivation" }, expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "s")
   public void testDependentScopedProducerFieldReturnsNonSerializableObjectForInjectionIntoInitializerParameterOfWebBeanWithPassivatingScopeFails()
   {
      getInstanceByType(Mikkeli_Broken.class).ping();
   }

   @Test(groups = { "contexts", "passivation" }, expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "t")
   public void testDependentScopedProducerFieldReturnsNonSerializableObjectForInjectionIntoProducerMethodParameterWithPassivatingScopeFails()
   {
      getInstanceByType(Violation2.class,BIG_LITERAL).ping();
   }
   
   @Test(groups = { "contexts", "passivation", "enterpriseBean", "integration" })
   //@SpecAssertion(section = "8.4", id = "a")
   @SpecAssertion(section = "unknown", id = "unknown")
   public void testEJBWebBeanCanDeclarePassivatingScope()
   {
      assert getBeans(TurkuLocal.class).size() == 1;
   }
   
   @Test(expectedExceptions = IllegalProductException.class)
   @SpecAssertion(section = "8.4", id = "cc")   
   public void testNonSerializableProducerFieldDeclaredPassivatingThrowsIllegalProductException()
   {
      getInstanceByType(HelsinkiNonSerializable.class).ping();
   }

   @Test(groups = { "contexts", "passivation" })
   @SpecAssertion(section = "8.4", id = "v")
   public void testIsSerializableOnBean()
   {
      assert !getBeans(Sysma.class).iterator().next().isSerializable();
      assert getBeans(Hyvinkaa.class).iterator().next().isSerializable();
   }
}
