package org.jboss.jsr299.tck.tests.event;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.Set;

import javax.enterprise.event.Observer;
import javax.enterprise.inject.AnnotationLiteral;
import javax.enterprise.inject.spi.BeanManager;

import org.hibernate.tck.annotations.SpecAssertion;
import org.hibernate.tck.annotations.SpecAssertions;
import org.jboss.jsr299.tck.AbstractJSR299Test;
import org.jboss.testharness.impl.packaging.Artifact;
import org.jboss.testharness.impl.packaging.jsr299.BeansXml;
import org.testng.annotations.Test;

/**
 * Event bus tests
 * 
 * @author Nicklas Karlsson
 * @author David Allen
 * 
 * Spec version: Public Release Draft 2
 */
@Artifact
@BeansXml("beans.xml")
public class EventTest extends AbstractJSR299Test
{
   private static final String FIRE_EVENT_METHOD_NAME = "fireEvent";
   private static final String RESOLVE_OBSERVERS_METHOD_NAME = "resolveObservers";
   
   public static class AnEventType
   {
   }

   public static class ATemplatedEventType<T>
   {
   }

   public static class AnObserver implements Observer<AnEventType>
   {
      public boolean wasNotified = false;

      public boolean notify(AnEventType event)
      {
         wasNotified = true;
         return false;
      }
   }

   public static class AListObserver implements Observer<ArrayList<String>>
   {
      public boolean wasNotified = false;

      public boolean notify(ArrayList<String> event)
      {
         wasNotified = true;
         return false;
      }
   }

   public static class AnotherListObserver implements Observer<ArrayList<Integer>>
   {
      public boolean wasNotified = false;

      public boolean notify(ArrayList<Integer> event)
      {
         wasNotified = true;
         return false;
      }
   }

   public static class AnObserverWithException implements Observer<AnEventType>
   {
      public boolean wasNotified = false;
      public RuntimeException theException = new RuntimeException("RE1");

      public boolean notify(AnEventType event)
      {
         wasNotified = true;
         throw theException;
      }
   }

   /**
    * An event binding type is a Java annotation defined as @Target({FIELD,
    * PARAMETER}) ~or @Target({METHOD, FIELD, PARAMETER, TYPE}) and
    * @Retention(RUNTIME)~. All event binding types must specify the
    * @BindingType meta-annotation.
    */
   @Test(groups = { "events" })
   @SpecAssertion(section = "10.1", id = "d")
   public void testEventBindingTypeTargetMostAndRuntime()
   {
      getCurrentManager().fireEvent("string event", new TameAnnotationLiteral());
   }

   /**
    * An event binding type is a Java annotation defined as ~@Target({FIELD,
    * PARAMETER}) or~ @Target({METHOD, FIELD, PARAMETER, TYPE}) ~and
    * @Retention(RUNTIME)~. All event binding types must specify the
    * @BindingType meta-annotation.
    */
   @Test(groups = { "events" })
   @SpecAssertion(section = "10.1", id = "e")
   public void testEventBindingTypeTargetsFieldParameterAndRuntime()
   {
      getCurrentManager().fireEvent("string event", new RoleBinding("Admin"));
   }

   /**
    * An event binding type is a Java annotation defined as ~@Target({FIELD,
    * PARAMETER}) or @Target({METHOD, FIELD, PARAMETER, TYPE}) and
    * ~@Retention(RUNTIME)
    */
   @Test(groups = { "events" }, expectedExceptions = { IllegalArgumentException.class } )
   @SpecAssertion(section = "10.1", id = "f")
   public void testEventBindingTypeNotRuntime()
   {
      getCurrentManager().fireEvent("string event", new AnnotationLiteral<BindingTypeNotRuntime>(){});
   }

   @Test(groups = { "events" }, expectedExceptions = { IllegalArgumentException.class } )
   @SpecAssertions({
      @SpecAssertion(section = "10.1", id = "g"),
      @SpecAssertion(section = "11.2.8", id = "e")
   })
   public void testManagerFireEventWithNonBindingAnnotationsFails()
   {
      getCurrentManager().fireEvent("string event", new AnnotationLiteral<NotABindingType>(){});
   }

   @Test(groups = { "events" }, expectedExceptions = { IllegalArgumentException.class })
   @SpecAssertions({
      @SpecAssertion(section = "10.1", id = "aa"),
      @SpecAssertion(section = "11.2.8", id = "c")
   })
   public void testManagerFireEventWithEventTypeParametersFails()
   {
      ATemplatedEventType<String> anEvent = new ATemplatedEventType<String>();
      getCurrentManager().fireEvent(anEvent);
   }

   @Test(groups = { "events" }, expectedExceptions = { IllegalArgumentException.class })
   @SpecAssertions({
      @SpecAssertion(section = "10.1", id = "aa"),
      @SpecAssertion(section = "11.2.8", id = "c")
   })
   public void testManagerFireEventWithEventTypeWildcardsFails()
   {
      ATemplatedEventType<?> anEventOnAnyType = new ATemplatedEventType<String>();
      getCurrentManager().fireEvent(anEventOnAnyType);
   }

   @Test(groups = { "events" }, expectedExceptions = { IllegalArgumentException.class })
   @SpecAssertion(section = "11.2.8", id = "d")
   public void testDuplicateBindingsToFireEventFails() throws Exception
   {
      getCurrentManager().fireEvent("string event", new TameAnnotationLiteral(), new TameAnnotationLiteral());
   }
   
   @Test(groups = { "events" })
   @SpecAssertions({
      @SpecAssertion(section = "11.2.8", id = "a"),
      @SpecAssertion(section = "11.2.8", id = "b")
   })
   public void testManagerInterfaceForFireEventMethod() throws Exception
   {
      assert BeanManager.class.getDeclaredMethod(FIRE_EVENT_METHOD_NAME, Object.class, Annotation[].class) != null;
   }

   /**
    * The Manager interface provides a method for firing events:
    */
   @Test(groups = { "events" })
   @SpecAssertion(section = "11.2.8", id = "a")
   public void testManagerFireEvent()
   {
      // First a simple event with no bindings is fired
      AnEventType anEvent = new AnEventType();
      getCurrentManager().fireEvent(anEvent);

      // Next an event with some event bindings is fired
      getCurrentManager().fireEvent(anEvent, new RoleBinding("Admin"));
   }

   @Test(groups = { "events" })
   @SpecAssertion(section = "10.3", id = "a")
   public void testObserversImplementObserverInterface()
   {
      Observer<String> observer = new Observer<String>()
      {
         public boolean notify(String event)
         {
            return false;
         }
      };
      getCurrentManager().addObserver(observer);
   }

   @Test(groups = { "events" })
   @SpecAssertions({
      @SpecAssertion(section = "11.2.9", id = "a"),
      @SpecAssertion(section = "11.2.9", id = "b")
   })
   public void testManagerResolveObserversSignature() throws Exception
   {
      assert getCurrentManager().getClass().getDeclaredMethod(RESOLVE_OBSERVERS_METHOD_NAME, Object.class, Annotation[].class) != null;
   }

   @Test(groups = { "events" })
   @SpecAssertions({
      @SpecAssertion(section = "10.5", id = "a"),
      @SpecAssertion(section = "10.5.2", id = "a"),
      @SpecAssertion(section = "12.3", id = "o")
   })
   public void testObserverMethodAutomaticallyRegistered()
   {
      assert !getCurrentManager().resolveObservers(new String()).isEmpty();
   }
   
   @Test(groups = { "events", "ri-broken" })
   @SpecAssertion(section = "12.3", id = "o")
   public void testObserverMethodNotAutomaticallyRegisteredForDisabledBeans()
   {
      Set<Observer<String>> observers = getCurrentManager().resolveObservers(new String());
      assert observers.size() == 2;
      for (Observer<String> observer : observers)
      {
         // an assertion error will be raised if an inappropriate observer is called
         observer.notify("fail if disable observer invoked");
      }
   }

   @Test(groups = { "events" })
   @SpecAssertion(section = "10.5", id = "e")
   public void testMultipleObserverMethodsForSameEventPermissible()
   {
      assert getCurrentManager().resolveObservers(new String()).size() > 1;
   }

   @Test(groups = { "events" })
   @SpecAssertion(section = "10.5", id = "f")
   public void testMultipleObserverMethodsOnBeanPermissible()
   {
      // Resolve the observers for types 1 and 2
      Set<Observer<ObservedType1>> resolvedStringObservers = getCurrentManager().resolveObservers(new ObservedType1());
      assert resolvedStringObservers.size() == 1;

      Set<Observer<ObservedType2>> resolvedIntegerObservers = getCurrentManager().resolveObservers(new ObservedType2());
      assert resolvedIntegerObservers.size() == 1;
   }

   @Test(groups = { "events", "ejb" })
   @SpecAssertion(section = "10.5", id = "d")
   public void testObserverMethodOnEnterpriseBeanIsBusinessMethodOrStatic()
   {
      Set<Observer<EJBEvent>> observers = getCurrentManager().resolveObservers(new EJBEvent());
      assert observers.size() == 2;
   }

   @Test(groups = { "events" })
   @SpecAssertion(section = "10.5.1", id = "b")
   public void testObserverMethodWithoutBindingTypesObservesEventsWithoutBindingTypes()
   {
      // Resolve registered observers with an event containing no binding types
      Set<Observer<SimpleEventType>> resolvedObservers = getCurrentManager().resolveObservers(new SimpleEventType());
      assert resolvedObservers.size() == 2;
   }

   @Test(groups = { "events" })
   @SpecAssertions( {
      @SpecAssertion(section = "10.5.2", id = "c"),
      @SpecAssertion(section = "10.2.1", id = "a"),
      @SpecAssertion(section = "10.2.2", id = "a")
   })
   public void testObserverMethodMayHaveMultipleBindingTypes()
   {
      // If we can resolve the observer with the two binding types, then it worked
      Set<Observer<MultiBindingEvent>> resolvedObservers = getCurrentManager().resolveObservers(new MultiBindingEvent(), new RoleBinding("Admin"), new TameAnnotationLiteral());
      assert resolvedObservers.size() == 2;
   }

   @Test(groups = { "events" })
   @SpecAssertions( {
      @SpecAssertion(section = "10.5.3", id = "a"),
      @SpecAssertion(section = "5.4.8", id = "c")
   })
   public void testObserverMethodReceivesInjectionsOnNonObservesParameters()
   {
      getCurrentManager().fireEvent("validate injected parameters");
   }

   @Test(groups = { "events" })
   @SpecAssertions( {
      @SpecAssertion(section = "10.5.4", id = "a"),
      @SpecAssertion(section = "5.4.8", id = "b")
   } )
   public void testConditionalObserver() throws Exception
   {
      RecluseSpider.reset();
      getCurrentManager().fireEvent(new ConditionalEvent());
      // Should not be notified since bean is not instantiated yet
      assert !RecluseSpider.isNotified();

      // Now instantiate the bean and fire another event
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            RecluseSpider bean = getInstanceByType(RecluseSpider.class);
            assert bean != null;
            // Must invoke a method to really create the instance
            assert !bean.isInstanceNotified();
            getCurrentManager().fireEvent(new ConditionalEvent());
            assert RecluseSpider.isNotified() && bean.isInstanceNotified();
         }

      }.run();
      
      RecluseSpider.reset();
   }
   
   @Test(groups = { "events" })
   @SpecAssertion(section = "10.6.1", id = "aa")
   public void testObserverMethodRegistration()
   {
      // Resolve registered observers with an event containing no binding types
      Set<Observer<SimpleEventType>> resolvedObservers = getCurrentManager().resolveObservers(new SimpleEventType());
      assert resolvedObservers.size() == 2;
   }

   @Test(groups = { "events" })
   @SpecAssertions({
      @SpecAssertion(section = "10.6.1", id = "b"),
      @SpecAssertion(section = "10.6.1", id = "h")
   })
   public void testObserverMethodCalledImmediately() throws Exception
   {
      StaticObserver.reset();
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            getCurrentManager().fireEvent(new Delivery());
            assert StaticObserver.isDeliveryReceived() && StaticObserver.getThreadObservingEvent().equals(Thread.currentThread());
         }
         
      }.run();
      StaticObserver.reset();
   }

   /**
    * FIXME the spec doesn't follow this exactly because technically it isn't supposed to use
    * the bean resolution alorithm to obtain the instance, which it does.
    */
   @Test(groups = { "events" })
   @SpecAssertions({
      @SpecAssertion(section = "10.5", id = "c"),
      @SpecAssertion(section = "5.4.8", id = "k")
   })
   public void testStaticObserverMethodInvoked() throws Exception
   {
      try
      {
         getCurrentConfiguration().getContexts().setInactive(getCurrentConfiguration().getContexts().getRequestContext());
         StaticObserver.reset();
         getCurrentManager().fireEvent(new Delivery());
         assert StaticObserver.isDeliveryReceived();
         StaticObserver.reset();
      }
      finally
      {
         getCurrentConfiguration().getContexts().setActive(getCurrentConfiguration().getContexts().getRequestContext());
      }
   }

   @Test(groups = { "events" })
   @SpecAssertions({
      @SpecAssertion(section = "4.3.2", id = "a"),
      @SpecAssertion(section = "5.4.8", id = "b")
   })
   public void testObserverCalledOnMostSpecializedInstance()
   {
      Shop.deliveryObservedBy = null;
      getCurrentManager().fireEvent(new Delivery());
      assert Shop.deliveryObservedBy.equals(FarmShop.class.getName());
   }

   @Test(groups = { "events" })
   @SpecAssertion(section = "5.4.8", id = "b")
   public void testObserverMethodInvokedOnReturnedInstanceFromContext()
   {
      RecluseSpider spider = getInstanceByType(RecluseSpider.class);
      spider.setWeb(new Web());
      getCurrentManager().fireEvent(new ConditionalEvent());
      assert spider.isInstanceNotified();
      assert spider.getWeb().getRings() == 1;
   }

   @Test(groups = { "events", "stub" })
   @SpecAssertion(section = "10.6.1", id = "p")
   public void testAsynchronousObserverThrowsExceptionIsLogged() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            getCurrentManager().fireEvent('a');
            Thread.sleep(200);
            //TODO How can we assert that the exception was logged?
         }
         
      }.run();
   }

   @Test(groups = { "events", "stub" })
   @SpecAssertion(section = "10.6.2", id = "a")
   public void testAsynchronousObserverMethodContexts()
   {
      assert false;
   }

   @Test(groups = { "events" }, expectedExceptions = { IllegalArgumentException.class })
   @SpecAssertion(section = "11.2.8", id = "d")
   public void testDuplicateBindingsToFireFails() throws Exception
   {
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            SweeWaxbill bean = getInstanceByType(SweeWaxbill.class);
            bean.methodThatFiresEvent();
         }

      }.run();
   }

   @Test(groups = { "events" }, expectedExceptions = { IllegalArgumentException.class })
   @SpecAssertion(section = "11.2.7", id = "d")
   public void testDuplicateBindingsToObservesFails() throws Exception
   {
      new RunInDependentContext()
      {

         @Override
         protected void execute() throws Exception
         {
            SweeWaxbill bean = getInstanceByType(SweeWaxbill.class);
            bean.methodThatRegistersObserver();
         }

      }.run();
   }

   // FIXME the assertion is being verified indirectly by verifying outcome
   // to verify directly, would need to mock BeanManager (very hard) or
   // intercept the call to BeanManager#fireEvent() (equally hard)
   @Test(groups = { "events" })
   @SpecAssertion(section = "10.4.1", id = "ed")
   public void testFireMethodCallsManagerFireWithEventObject()
   {
      BullTerrier.reset();
      getInstanceByType(DogWhisperer.class).issueTamingCommand();
      assert BullTerrier.isMultiBindingEventObserved();
      assert BullTerrier.isSingleBindingEventObserved();
      BullTerrier.reset();
   }

   @Test(groups = { "events", "underInvestigation" })
   @SpecAssertion(section = "11.2.7", id = "a")
   public void testObserverMethodCallsManagerAddObserverWithObserverObject()
   {
      assert false;
   }

   @Test(groups = { "events", "underInvestigation" })
   @SpecAssertion(section = "10.4.1", id = "ee")
   public void testObserverMethodCallsManagerAddObserverWithAllBindingAnnotationsExceptFires()
   {
      assert false;
   }

   @Test(groups = { "events" }, expectedExceptions = IllegalArgumentException.class)
   @SpecAssertion(section = "11.2.9", id = "c")
   public void testEventObjectContainsTypeVariablesWhenResolvingFails()
   {
      eventObjectContainsTypeVariables(new ArrayList<String>());
   }

   private <E> void eventObjectContainsTypeVariables(ArrayList<E> eventToFire)
   {
      getCurrentManager().resolveObservers(eventToFire);
   }

   @Test(groups = { "events", "ri-broken" })
   @SpecAssertion(section = "unknown", id = "unknown")
   public void testEventObjectContainsWildcardsWhenResolvingFails()
   {
      eventObjectContainsWildcards(new ArrayList<String>());
   }

   private void eventObjectContainsWildcards(ArrayList<?> eventToFire)
   {
      // FIXME create a valid observer and assert it is found
      getCurrentManager().resolveObservers(eventToFire);
   }

   
   @Test(groups = { "events" })
   @SpecAssertions( {
      @SpecAssertion(section = "10.2.2", id = "b"),
      @SpecAssertion(section = "10.2.2", id = "c")
   } )
   public void testObserverMethodNotifiedWhenBindingsMatch()
   {
      getCurrentManager().fireEvent(new MultiBindingEvent(), new RoleBinding("Admin"), new TameAnnotationLiteral());
      assert BullTerrier.isMultiBindingEventObserved();
      assert BullTerrier.isSingleBindingEventObserved();
   }

   @Test(groups = { "events", "stub" })
   @SpecAssertions( { @SpecAssertion(section = "10.7", id = "a") })
   public void testEventTypeMappedToJmsTopic()
   {
      assert false;
   }

   @Test(groups = { "events", "stub" })
   @SpecAssertions( { @SpecAssertion(section = "10.7", id = "b") })
   public void testEventMappingIsObserverMethodDeclaredByInterface()
   {
      assert false;
   }

   @Test(groups = { "events", "stub" })
   @SpecAssertions( { @SpecAssertion(section = "10.7", id = "c") })
   public void testEventParameterSpecifiesMappedEventTypeAndBindings()
   {
      assert false;
   }
   
   @Test(groups = { "events", "stub" })
   @SpecAssertions( { @SpecAssertion(section = "10.7", id = "d") })
   public void testMessageDestinationInjectedIsMappedTopic()
   {
      assert false;
   }
   
   @Test(groups = { "events", "stub" })
   @SpecAssertions( { @SpecAssertion(section = "10.7", id = "e") })
   public void testEventMappingMemberOfAnyInterface()
   {
      assert false;
   }
   
   @Test(groups = { "events", "stub" })
   @SpecAssertions( { @SpecAssertion(section = "10.7", id = "f") })
   public void testObserversOfJmsTopicMustBeAsynchronous()
   {
      assert false;
   }

   @Test(groups = { "events", "stub" })
   @SpecAssertions( { @SpecAssertion(section = "10.7", id = "g") })
   public void testObserversNotifiedOfEventFromJmsTopic()
   {
      assert false;
   }

   @Test(groups = { "events", "stub" })
   @SpecAssertions( { @SpecAssertion(section = "10.7", id = "g") })
   public void testEventsDistributedToProcessesWithSameTypeAndJmsTopic()
   {
      assert false;
   }

   /**
    * By default, Java implementation reuse is assumed. In this case, the
    * producer, disposal and observer methods of the first bean are not
    * inherited by the second bean.
    * 
    * @throws Exception
    */
   @Test(groups = { "events", "inheritance" })
   @SpecAssertion(section = "4.2", id = "dc")
   public void testNonStaticObserverMethodNotInherited() throws Exception
   {
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            Egg egg = new Egg();
            Set<Observer<Egg>> observers = getCurrentManager().resolveObservers(egg);
            assert observers.size() == 1;

            // Notify the observer so we can confirm that it
            // is a method only on Farmer, and not LazyFarmer
            observers.iterator().next().notify(egg);
            assert egg.getClassesVisited().size() == 1;
            assert egg.getClassesVisited().iterator().next().equals(Farmer.class);
         }

      }.run();
   }
   
   @Test(groups = { "events", "inheritance" })
   @SpecAssertion(section = "4.2", id = "di")
   public void testNonStaticObserverMethodNotIndirectlyInherited() throws Exception
   {
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            StockPrice price = new StockPrice();
            Set<Observer<StockPrice>> observers = getCurrentManager().resolveObservers(price);
            assert observers.size() == 1;

            // Notify the observer so we can confirm that it
            // is a method only on StockWatcher, and not IntermediateStockWatcher
            // or IndirectStockWatcher
            observers.iterator().next().notify(price);
            assert price.getClassesVisited().size() == 1;
            assert price.getClassesVisited().iterator().next().equals(StockWatcher.class);
         }

      }.run();
   }

}
