package org.jboss.jsr299.tck.tests.implementation.enterprise.lifecycle;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import javax.enterprise.context.RequestScoped;
import javax.enterprise.context.spi.Context;
import javax.enterprise.context.spi.CreationalContext;
import javax.enterprise.inject.AnnotationLiteral;
import javax.enterprise.inject.spi.Bean;

import org.hibernate.tck.annotations.SpecAssertion;
import org.hibernate.tck.annotations.SpecAssertions;
import org.jboss.jsr299.tck.AbstractJSR299Test;
import org.jboss.testharness.impl.packaging.Artifact;
import org.jboss.testharness.impl.packaging.IntegrationTest;
import org.jboss.testharness.impl.packaging.Packaging;
import org.jboss.testharness.impl.packaging.PackagingType;
import org.testng.annotations.Test;

/**
 * Sections
 * 
 * 6.5. Lifecycle of stateful session beans
 * 6.6. Lifecycle of stateless session and singleton beans
 * 6.11. Lifecycle of EJBs
 * 
 * Mostly overlapping with other tests...
 * 
 * @author Nicklas Karlsson
 * @author David Allen
 * 
 * Spec version: 20090519
 * 
 */
@Artifact
@Packaging(PackagingType.EAR)
@IntegrationTest
public class EnterpriseBeanLifecycleTest extends AbstractJSR299Test
{

   @Test(groups = {"enterpriseBeans", "clientProxy", "lifecycle", "integration" })
   @SpecAssertions( {
      @SpecAssertion(section = "3.3.5", id = "a"),
      @SpecAssertion(section = "7.3", id = "aa"),
      @SpecAssertion(section = "7.3", id = "bb")
   })
   public void testCreateSFSB() throws Exception
   {
      GrossStadt frankfurt = getInstanceByType(GrossStadt.class);
      Bean<KleinStadt> stadtBean = getBeans(KleinStadt.class).iterator().next();
      assert stadtBean != null : "Expected a bean for stateful session bean Kassel";
      KleinStadt stadtInstance = getInstanceByType(KleinStadt.class, new AnnotationLiteral<Important>() {});
      assert stadtInstance != null : "Expected instance to be created by container";
      assert frankfurt.isKleinStadtCreated() : "PostConstruct should be invoked when bean instance is created";
      frankfurt.resetCreatedFlags();
      
      // Create a second one to make sure create always does create a new session bean
      KleinStadt anotherStadtInstance = getInstanceByType(KleinStadt.class, new AnnotationLiteral<Important>() {});
      assert anotherStadtInstance != null : "Expected second instance of session bean";
      assert frankfurt.isKleinStadtCreated();
      assert anotherStadtInstance != stadtInstance : "create() should not return same bean as before";
      
      // Verify that the instance returned is a proxy by checking for all local interfaces
      Set<Class> interfaces = new HashSet<Class>(Arrays.asList(stadtInstance.getClass().getInterfaces()));
      assert interfaces.contains(KleinStadt.class);
      assert interfaces.contains(SchoeneStadt.class);
      
   }
   
   @Test(groups = { "ri-broken", "enterpriseBeans", "clientProxy", "lifecycle", "integration" })
   @SpecAssertions( {
      @SpecAssertion(section = "6.6.2", id = "a")
   })
   public void testSerializeSFSB() throws Exception
   {
      KleinStadt stadtInstance = getInstanceByType(KleinStadt.class, new AnnotationLiteral<Important>() {});

      byte[] bytes = serialize(stadtInstance);
      Object object = deserialize(bytes);
      stadtInstance = (KleinStadt) object;
      assert getCurrentConfiguration().getBeans().isProxy(stadtInstance);
      
   }

   @Test(groups = { "enterpriseBeans", "clientProxy", "lifecycle", "integration" })
   @SpecAssertions({
      @SpecAssertion(section = "7.3", id = "bc")
   })
   public void testDestroyRemovesSFSB() throws Exception
   {
      GrossStadt frankfurt = getInstanceByType(GrossStadt.class);
      Bean<KleinStadt> stadtBean = getBeans(KleinStadt.class).iterator().next();
      assert stadtBean != null : "Expected a bean for stateful session bean Kassel";
      Context requestContext = getCurrentManager().getContext(RequestScoped.class);
      CreationalContext<KleinStadt> creationalContext = getCurrentManager().createCreationalContext(stadtBean);
      KleinStadt kassel = stadtBean.create(creationalContext);
      kassel.ping();
      stadtBean.destroy(kassel, creationalContext);
      
      assert frankfurt.isKleinStadtDestroyed() : "Expected SFSB bean to be destroyed";
      kassel = requestContext.get(stadtBean);
      assert kassel == null : "SFSB bean should not exist after being destroyed";
      //frankfurt.dispose();
   }

   @Test(groups = { "enterpriseBeans", "lifecycle", "integration" })
   @SpecAssertions({
      @SpecAssertion(section = "7.3", id = "bc"),
      @SpecAssertion(section = "3.3.1", id = "dba")
   })
   public void testRemovedEjbIgnored()
   {
      KleinStadt stadtInstance = getInstanceByType(KleinStadt.class, new AnnotationLiteral<Important>() {});
      assert stadtInstance != null : "Expected instance to be created by container";
      stadtInstance.setName("Kassel-Wilhelmshoehe");
      stadtInstance.zustandVergessen();
      
      // Now make sure that the container does not return this instance again
      KleinStadt newStadtInstance = getInstanceByType(KleinStadt.class);
      assert newStadtInstance != null : "Failed to get SFSB instance the second time";
      assert !"Kassel-Wilhelmshoehe".equals(newStadtInstance.getName()) : "The destroyed SFSB was not ignored";
   }

   @Test(groups = { "enterpriseBeans", "lifecycle", "integration" })
   @SpecAssertion(section = "7.4", id = "a")
   public void testCreateSLSB()
   {
      Bean<NeueStadt> stadtBean = getBeans(NeueStadt.class).iterator().next();
      assert stadtBean != null : "Expected a bean for stateful session bean Kassel";
      CreationalContext<NeueStadt> creationalContext = getCurrentManager().createCreationalContext(stadtBean);
      NeueStadt stadtInstance = stadtBean.create(creationalContext);
      assert stadtInstance != null : "Expected instance to be created by container";

      // Verify that the instance returned is a proxy by checking for all local interfaces
      Set<Class> interfaces = new HashSet<Class>(Arrays.asList(stadtInstance.getClass().getInterfaces()));
      assert interfaces.contains(NeueStadt.class);
      assert interfaces.contains(GeschichtslosStadt.class);
   }

   @Test(groups = { "enterpriseBeans", "lifecycle", "integration" })
   @SpecAssertion(section = "3.9.2", id = "b")
   public void testInitializerMethodsCalledWithCurrentParameterValues()
   {
      AlteStadt alteStadt = this.getCurrentConfiguration().getBeans().getEnterpriseBean(Mainz.class, AlteStadt.class);
      assert alteStadt != null : "Could not find the AlteStadt EJB";
      assert alteStadt.getAnotherPlaceOfInterest() != null;
   }

   @Test(groups = { "enterpriseBeans", "lifecycle", "ri-broken" })
   @SpecAssertion(section = "6.11", id = "f")
   public void testDependentObjectsDestroyed()
   {
      UniStadt marburg = getInstanceByType(UniStadt.class);
      assert marburg != null : "Couldn't find the main SFSB";
      Bean<UniStadt> uniStadtBean = getBeans(UniStadt.class).iterator().next();
      CreationalContext<UniStadt> creationalContext = getCurrentManager().createCreationalContext(uniStadtBean);      
      uniStadtBean.destroy(marburg, creationalContext);
      GrossStadt frankfurt = getInstanceByType(GrossStadt.class);
      assert frankfurt.isSchlossDestroyed();
   }

   @Test
   @SpecAssertion(section = "4.2", id = "bab")
   public void testDirectSubClassInheritsPostConstructOnSuperclass() throws Exception
   {
      OrderProcessor.postConstructCalled = false;
      assert getBeans(DirectOrderProcessorLocal.class).size() == 1;
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            getInstanceByType(DirectOrderProcessorLocal.class).order();
         }
      }.run();
      assert OrderProcessor.postConstructCalled;
   }
   
   @Test
   @SpecAssertion(section = "4.2", id = "bad")
   public void testIndirectSubClassInheritsPostConstructOnSuperclass() throws Exception
   {
      OrderProcessor.postConstructCalled = false;
      assert getBeans(OrderProcessorLocal.class).size() == 1;
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            getInstanceByType(OrderProcessorLocal.class).order();
         }
      }.run();
      assert OrderProcessor.postConstructCalled;
   }
   
   @Test(groups = "ri-broken")
   @SpecAssertion(section = "4.2", id = "bbb")
   public void testSubClassInheritsPreDestroyOnSuperclass() throws Exception
   {
      OrderProcessor.preDestroyCalled = false;
      assert getBeans(DirectOrderProcessorLocal.class).size() == 1;
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            Bean<DirectOrderProcessorLocal> bean = getBeans(DirectOrderProcessorLocal.class).iterator().next();
            CreationalContext<DirectOrderProcessorLocal> creationalContext = getCurrentManager().createCreationalContext(bean);
            DirectOrderProcessorLocal instance = getInstanceByType(DirectOrderProcessorLocal.class);
            bean.destroy(instance, creationalContext);
         }
      }.run();
      assert OrderProcessor.preDestroyCalled;
   }
   
   @Test(groups = "ri-broken")
   @SpecAssertion(section = "4.2", id = "bbd")
   public void testIndirectSubClassInheritsPreDestroyOnSuperclass() throws Exception
   {
      OrderProcessor.preDestroyCalled = false;
      assert getBeans(IndirectOrderProcessor.class).size() == 1;
      new RunInDependentContext()
      {
         @Override
         protected void execute() throws Exception
         {
            Bean<IndirectOrderProcessor> bean = getBeans(IndirectOrderProcessor.class).iterator().next();
            CreationalContext<IndirectOrderProcessor> creationalContext = getCurrentManager().createCreationalContext(bean);
            IndirectOrderProcessor instance = getInstanceByType(IndirectOrderProcessor.class);
            bean.destroy(instance, creationalContext);
         }
      }.run();
      assert OrderProcessor.preDestroyCalled;
   }
}
